/* ====================================================================
 *  Copyright 2004 Edward Rudd and Paul Querna
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Some code based off of Subversion:
 *     Copyright (c) CollabNet.  All rights reserved.
 *
 */

#include "http_core.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_log.h"
#include "svn_repos.h"
#include "svn_path.h"
#include "svn_config.h"
#include "svn_string.h"
#include "mod_svn_view_private.h"

/*
 * This code is *completely* based off of mod_authz_svn:
 *   http://svn.collab.net/repos/svn/trunk/subversion/mod_authz_svn/mod_authz_svn.c
 *
 * If there are vulnerabilities found in authz_svn, you MUST update this file!
 *
 * It would be nice if these functions where part of a core subversion lib. Oh Well. 
 * Copy and pasting we a`go.
 *
 */

enum {
    AUTHZ_SVN_NONE = 0,
    AUTHZ_SVN_READ = 1,
    AUTHZ_SVN_WRITE = 2,
    AUTHZ_SVN_RECURSIVE = 4
};

struct parse_authz_baton {
    apr_pool_t *pool;
    svn_config_t *config;
    const char *user;
    int allow;
    int deny;

    int required_access;
    const char *repos_path;
    const char *qualified_repos_path;

    int access;
};

/*
 * Access checking
 */

static int group_contains_user(svn_config_t *cfg,
    const char *group, const char *user, apr_pool_t *pool)
{
    const char *value;
    apr_array_header_t *list;
    int i;

    svn_config_get(cfg, &value, "groups", group, "");
    list = svn_cstring_split(value, ",", TRUE, pool);

    for (i = 0; i < list->nelts; i++) {
       const char *group_user = APR_ARRAY_IDX(list, i, char *);
       if (!strcmp(user, group_user))
           return 1;
    }

    return 0;
}

static svn_boolean_t parse_authz_line(const char *name, const char *value,
                                      void *baton)
{
    struct parse_authz_baton *b = baton;

    if (strcmp(name, "*")) {
        if (!b->user) {
            return TRUE;
        }

        if (*name == '@') {
            if (!group_contains_user(b->config, &name[1], b->user, b->pool))
                return TRUE;
        }
        else if (strcmp(name, b->user)) {
            return TRUE;
        }
    }

    if (ap_strchr_c(value, 'r')) {
        b->allow |= AUTHZ_SVN_READ;
    }
    else {
        b->deny |= AUTHZ_SVN_READ;
    }

    if (ap_strchr_c(value, 'w')) {
        b->allow |= AUTHZ_SVN_WRITE;
    }
    else {
        b->deny |= AUTHZ_SVN_WRITE;
    }

    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, b->pool,
                  "%s = %s => allow = %i, deny = %i",
                  name, value, b->allow, b->deny);

    return TRUE;
}

/*
 * Return TRUE when ACCESS has been determined.
 */
static int parse_authz_lines(svn_config_t *cfg,
                             const char *repos_name, const char *repos_path,
                             const char *user,
                             int required_access, int *granted_access,
                             apr_pool_t *pool)
{
    const char *qualified_repos_path;
    struct parse_authz_baton baton = { 0 };

    baton.pool = pool;
    baton.config = cfg;
    baton.user = user;

    /* First try repos specific */
    qualified_repos_path = apr_pstrcat(pool, repos_name, ":", repos_path,
                                       NULL);
    svn_config_enumerate(cfg, qualified_repos_path,
                         parse_authz_line, &baton);
    *granted_access = !(baton.deny & required_access)
                      || (baton.allow & required_access);

    if ((baton.deny & required_access)
        || (baton.allow & required_access))
        return TRUE;

    svn_config_enumerate(cfg, repos_path,
                         parse_authz_line, &baton);
    *granted_access = !(baton.deny & required_access)
                      || (baton.allow & required_access);

    return (baton.deny & required_access)
           || (baton.allow & required_access);
}

static svn_boolean_t parse_authz_section(const char *section_name,
                                         void *baton)
{
  struct parse_authz_baton *b = baton;
  int conclusive;

  if (strncmp(section_name, b->qualified_repos_path,
              strlen(b->qualified_repos_path))
      && strncmp(section_name, b->repos_path,
                 strlen(b->repos_path))) {
      /* No match, move on to the next section. */
      return TRUE;
  }

  b->allow = b->deny = 0;
  svn_config_enumerate(b->config, section_name,
                       parse_authz_line, b);

  conclusive = (b->deny & b->required_access)
               || (b->allow & b->required_access);

  b->access = !(b->deny & b->required_access)
              || (b->allow & b->required_access)
              || !conclusive;
  
  /* If access isn't denied, move on to check the next section. */
  return b->access;
}

static int parse_authz_sections(svn_config_t *cfg,
                                const char *repos_name, const char *repos_path,
                                const char *user,
                                int required_access,
                                apr_pool_t *pool)
{
    struct parse_authz_baton baton = { 0 };

    baton.pool = pool;
    baton.config = cfg;
    baton.user = user;
    baton.required_access = required_access;
    baton.repos_path = repos_path;
    baton.qualified_repos_path = apr_pstrcat(pool, repos_name, ":",
                                             repos_path, NULL);
    
    baton.access = 1; /* Allow by default */
    svn_config_enumerate_sections(cfg, parse_authz_section, &baton);

    return baton.access;
}


static int check_access(svn_config_t *cfg, const char *repos_name,
                        const char *repos_path, const char *user,
                        int required_access, apr_pool_t *pool)
{
    const char *base_name;
    const char *original_repos_path = repos_path;
    int granted_access;

    if (!repos_path) {
        /* XXX: Check if the user has 'required_access' _anywhere_ in the
         * XXX: repository.  For now, make this always succeed, until
         * XXX: we come up with a good way of figuring this out.
         */
        return 1;
    }

    base_name = repos_path;
    while (!parse_authz_lines(cfg, repos_name, repos_path,
                              user, required_access, &granted_access,
                              pool)) {
        if (base_name[0] == '/' && base_name[1] == '\0') {
            /* By default, deny access */
            return 0;
        }

        svn_path_split(repos_path, &repos_path, &base_name, pool);
    }

    if (granted_access && (required_access & AUTHZ_SVN_RECURSIVE) != 0) {
        /* Check access on entries below the current repos path */
        granted_access = parse_authz_sections(cfg,
                                              repos_name, original_repos_path,
                                              user, required_access,
                                              pool);
    }

    return granted_access;
}

/* TODO: Should this function be part of a public API? */

static int sv_check_access(svn_repo_rec* svn, const char* path_to_check, 
                           const char* username, int required_access, apr_pool_t* pool) 
{
    /* If a AuthzFile was not defined, 
       we will default to ENABLE access */
    if(svn->authz_file == NULL) {
        return 1;
    }

    if(svn->authz == NULL) {
        svn_error_t *err;
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, svn->r,
                      "reading authz file: %s", svn->authz_file);

        err = svn_config_read(&svn->authz, svn->authz_file, FALSE,
                                  svn->p);
        if (err != SVN_NO_ERROR) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, err->apr_err, svn->r,
                          "svn_view/authz: %s", err->message);

            /* If there was some error reading the configuration file,
               deny access! */
            return 0;
        }        
    }

    return check_access(svn->authz, svn->reponame, path_to_check,
                        username, required_access, pool); 
}

APR_DECLARE( int ) sv_authz_check(svn_repo_rec * svn, 
                                const char* path_to_check) 
{
    return sv_check_access(svn, path_to_check, svn->r->user, AUTHZ_SVN_READ, svn->p);
}


APR_DECLARE( svn_error_t *) sv_authz_callback (svn_boolean_t *allowed,
                                              svn_fs_root_t *root,
                                              const char *path,
                                              void *baton,
                                              apr_pool_t *pool) 
{
    svn_repo_rec* svn = baton;
    *allowed = sv_check_access(svn, path, svn->r->user, AUTHZ_SVN_READ, pool);
    return SVN_NO_ERROR;
}
