/* ====================================================================
 *  Copyright 2004 Edward Rudd and Paul Querna
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Some code based off of Subversion:
 *     Copyright (c) CollabNet.  All rights reserved.
 *
 */

/** 
 * mod_svn_view
 *  - Browse Subversion Repositories with an Apache Module
 */

/**
 *  Prerequisites:
 *    - An XSLT Output Filter
 *        - Tested with mod_transform: http://www.outoforder.cc/
 *    - Local Subversion Repositories
 * 
 *  Example Configuration:
 *    <Directory "/home/chip/work/mod_svn_view/trunk/themes">
 *        Order allow,deny
 *        Allow from All
 *        Deny From none
 *    </Directory>
 *
 *    <Location "/svn-view">
 *        SVNView on
 *        SVNViewRepository test /test/repos/test
 *        SVNViewRepository apache-2.0 /test/repos/httpd-2.0
 *        SVNViewRepository svn-mirror /test/repos/subversion
 *        SVNViewAuthzSvnFile /test/trunk/svn-authz
 *        SVNViewDefaultXSLT /home/chip/work/mod_svn_view/trunk/xslt/default.xsl
 *        SVNViewTheme blueview
 *        SVNViewThemesDir /home/chip/work/mod_svn_view/trunk/themes
 *    </Location>
 *
 *  TODO:
 *     - Public Website and Logo (based off of SVN Logo?)
 *         - Why should you use this over other programs?
 *     - Test on win32
 *     - Add Authz Checks on all Accesses.
 *
 */

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
/* to get ap_process_request. TODO: Why is this in Core Private? */
#define CORE_PRIVATE
#include "http_request.h"

#include "apr_tables.h"
#include "apr_hash.h"
#include "svn_repos.h"
#include "svn_path.h"
#include "svn_utf.h"
#include "svn_time.h"
#include "svn_props.h"
#include "mod_svn_view_private.h"

#define SVN_VIEW_VERSION VERSION

/* Content Type sent for IE/non cool browsers */
#define SV_COMPAT_CONTENT_TYPE "text/html"
/* Content Type for Browsers that Like XHTML */
#define SV_XHTML_CONTENT_TYPE "application/xhtml+xml"
/* Content Type for when the Query String has show_xml=1 */
#define SV_XML_CONTENT_TYPE "text/xml"
/* Name of the Output Filter that we use */
/* TODO: Make this a Config Option w/ Default Values... */
#define SV_XSLT_FILTER_NAME "XSLT"
#define SV_HIGHLIGHT_FILTER_NAME "HIGHLIGHT"

/* Properties that are used per-repo to set special values */
#define SV_PROP_PREFIX "svn-view:"
#define SV_PROP_CSS     SV_PROP_PREFIX "csspath"
#define SV_PROP_XSLT    SV_PROP_PREFIX "xsltpath"
#define SV_PROP_DESC    SV_PROP_PREFIX "description"

module AP_MODULE_DECLARE_DATA svn_view_module;

/* Apache directory configuration structure */
typedef struct sv_dir_config_rec sv_dir_config_rec;
struct sv_dir_config_rec {
    const char *sv_root;
    apr_hash_t *repos;
    const char *authzpath;
    const char *xsltpath;
    const char *theme_dir;
    const char *theme_name;
    const char* parent_path;
    const char* enabled;
};

/* action Handler function structure */
typedef struct sv_action_handler_st sv_action_handler_st;
struct sv_action_handler_st {
    sv_action_handler *func;
    int states;
    void *data;
};

typedef struct sv_history_baton_st sv_history_baton_st;
struct sv_history_baton_st {
    svn_repo_rec *svn;
    int num_left;
    int show_files;
};

apr_hash_t *sv_action_hash;

#define sv_action_get(action) \
    apr_hash_get(sv_action_hash, action, APR_HASH_KEY_STRING)

#define sv_action_call(handler, svn, parms) \
    handler->func(svn, parms, handler->data)

/* Action Functions */
static int sv_list_repos(svn_repo_rec * svn, apr_table_t * parms, void *data);
static int sv_view_dir(svn_repo_rec * svn, apr_table_t * parms, void *data);
static int sv_view_history(svn_repo_rec * svn,
    apr_table_t * parms, void *data);
static int sv_view_file(svn_repo_rec * svn, apr_table_t * parms, void *data);
static int sv_view_inline(svn_repo_rec * svn, apr_table_t * parms, void *data);
static int sv_view_commit(svn_repo_rec * svn, apr_table_t * parms, void *data);

/* Utility Functions */
static char *sv_build_query_string(apr_pool_t * p, apr_table_t * args)
    __attribute__ ((unused));
static void sv_fill_query_string_table(request_rec * r, apr_table_t * parms);
static int sv_show_props(request_rec * r, svn_repo_rec * svn,
    apr_hash_t * prop_list);

static svn_error_t *sv_print_changed_tree(svn_repos_node_t * node,
    const char *path,
    svn_repo_rec* svn);

static svn_error_t *sv_show_revision(void *p_baton, const char
    *path, svn_revnum_t revision, apr_pool_t * pool);

/* Repository utility functions */
static int sv_repo_open(request_rec * r, svn_repo_rec * svn);
static int sv_repo_getprops(request_rec * r, svn_repo_rec * svn,
    const char *svn_path, apr_hash_t ** prop_list);
static int sv_repo_setrev(request_rec * r, svn_repo_rec * svn);

static int sv_viewable_content_type(svn_string_t * mime_type);

/* Output Function */
static void sv_xml_header(apr_table_t * get_table,
    svn_repo_rec * svn, const char *action);
static void sv_xml_footer(request_rec * r);

/* TODO: Would be nice to remove this. */
#define SVN_ERROR_X(file, line, err,msg)  \
   if (err != SVN_NO_ERROR) { \
        char* m = apr_pstrcat(r->pool, "[%s:%d] ", msg, NULL); \
        ap_log_rerror(APLOG_MARK, APLOG_ERR, err->apr_err, r, \
            m, file, line, err->message); \
        return HTTP_INTERNAL_SERVER_ERROR; \
    }
#define SVN_ERROR(err,msg)  SVN_ERROR_X(__FILE__, __LINE__, err, msg)

/* Main Handler */
static int mod_svn_view_handler(request_rec * r)
{
    apr_pool_t *p;
    apr_table_t *get_table;
    const char *action, *svn_rev;
    int list_repo_detail = 0;
    int ret = OK;
    svn_error_t *err;
    svn_repo_rec *svn;
    sv_action_handler_st *action_handler;
    sv_dir_config_rec *dc = ap_get_module_config(r->per_dir_config,
        &svn_view_module);

    if (r->method_number != M_GET) {
        return DECLINED;
    }
    
    if (strcmp(r->handler, "svn-view") && strcmp(r->handler, "dav-handler")) {
        return DECLINED;
    }

    if (!dc->enabled) {
        return DECLINED;
    }

    if (strcmp(dc->enabled, "on") != 0) {
        return DECLINED;
    }

    /* TODO: Make an apr_table of XSLT per repo?? */
    if (!dc->xsltpath) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
            "XSLT Path not set for mod_svn_view");
        return HTTP_INTERNAL_SERVER_ERROR;
    }
    
    /* Create a Subpool for Subversion to mess up */
    apr_pool_create(&p, r->pool);
    svn = apr_pcalloc(p, sizeof(*svn));
    svn->p = p;
    /* So the request_rec is always available. */
    svn->r = r;
    svn->sv_root = dc->sv_root;

    svn->authz_file = NULL;
    svn->authz = NULL;
    svn->repos = apr_hash_copy(p, dc->repos);
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  "Parent Path: '%s'",  dc->parent_path);
    if (dc->parent_path) {
        apr_dir_t *dir;
        apr_status_t rv;
        rv = apr_dir_open(&dir, dc->parent_path, p);
        if (rv == APR_SUCCESS) {
            apr_finfo_t finfo;
            while (apr_dir_read(&finfo, APR_FINFO_TYPE|APR_FINFO_NAME, dir) == 
                   APR_SUCCESS) {

                if (finfo.filetype == APR_DIR) {
                    if (strncmp(".", finfo.name, 1)) {
                        apr_hash_set(svn->repos, finfo.name, APR_HASH_KEY_STRING,
                                     apr_pstrcat(p, dc->parent_path, "/", finfo.name, NULL));
                    }
                }
            }
            apr_dir_close(dir);
        }
        else {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
                          "Unable to access Parent Path: '%s'",  dc->parent_path);
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }
    
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
        "Path_Info: %s", r->path_info);

    /*
     * parse and store query_string 
     */
    get_table = apr_table_make(svn->p, 2);
    sv_fill_query_string_table(r, get_table);
    action = apr_table_get(get_table, "action");

    /*
     * redirect to correct Repo URL from drop down combo box 
     */
    if ((action != NULL) && strcmp(action, "redirect") == 0) {
        const char *repo = apr_table_get(get_table, "repo");

        if (repo) {
            const char *repo_url =
                apr_pstrcat(r->pool, dc->sv_root, "/", repo, "/", NULL);
            apr_table_setn(r->headers_out, "Location",
                ap_construct_url(r->pool, repo_url, r));
            return HTTP_MOVED_TEMPORARILY;
        }
        else {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                "Something bad happened no repo");
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }

    /*
     * parse out Repository from path_info 
     */
    if (strlen(r->path_info) > 0) {
        int had_slash, len1;
        char* uri;

        uri = ap_strchr((r->path_info + 1), '/');
        svn->reponame =
            apr_pstrndup(r->pool, (r->path_info + 1),
            uri - (r->path_info + 1));

        /* remove duplicate slashes, and make sure URI has no trailing '/' */
        if(uri) {
            ap_no2slash(uri);
            len1 = strlen(uri);
            had_slash = (len1 > 0 && uri[len1 - 1] == '/');

            if (len1 > 1 && had_slash) {
                uri[len1 - 1] = '\0';
            }

            if (*uri == '\0') {
                uri = "/";
            }
            else if (*uri != '/') {
                --uri;
            }
        }
        else {
            uri = "/";
        }

        svn->path = apr_pstrdup(r->pool, uri);
        
        /* TODO: check for valid repo */
        if (strlen(svn->reponame) == 0) {
            svn->reponame = NULL;
        }
        else if(strncmp(svn->reponame, "*theme*", 7) == 0) {
            request_rec* sr;
            /* Special Case: A special 'theme' URL that
                             we re-map into another directory to
                             keep the XSL and configuration simple.
             */
            sr = ap_sub_req_lookup_file(apr_pstrcat(r->pool, dc->theme_dir, "/",
                                                    dc->theme_name, svn->path, NULL), 
                                        r, r->output_filters);
            if (sr->status != HTTP_OK) {
                ret = sr->status;
                ap_destroy_sub_req(sr);
                return ret;
            }
            /* TODO: This is sort of hacky, Im sure there is a better way. */
            ap_internal_fast_redirect(sr, r);
            ap_process_request(sr);
            return OK;
        }
        else {

            svn->repopath = apr_hash_get(svn->repos, svn->reponame,
                APR_HASH_KEY_STRING);
            if (!svn->repopath) {
                /* TODO: should we output a nice error message here? */
                ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
                    "Invalid Repository Requested: %s", svn->reponame);
                return HTTP_NOT_FOUND;
            }
        }
    }

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
        "Repo: %s, Path: %s", svn->reponame, svn->path);
    
    if (svn->path && strncmp(svn->path, "/!svn/", 6) == 0) {
        return DECLINED;
    }
    
    if (!svn->reponame) {
        action = "list_repos";
        list_repo_detail = 1;
        svn->xslt = dc->xsltpath;
    }
    else {
        apr_hash_t *prop_list;
        svn_string_t *value;
        svn_fs_root_t *rev_root;

        svn->authz_file = dc->authzpath;

        if(!sv_authz_check(svn, svn->path)) {
            return HTTP_FORBIDDEN;
        }

        /*
         * Open up the repository. 
         */
        if (OK != (ret = sv_repo_open(r, svn))) {
            return ret;
        }

        if ((svn_rev = apr_table_get(get_table, "rev"))) {
            svn->request_rev = apr_atoi64(svn_rev);
            sv_repo_setrev(r, svn);
        }
        else {
            svn->request_rev = SVN_INVALID_REVNUM;
            sv_repo_setrev(r, svn);     /* -1 sets revision to latest */
        }

        svn_fs_check_path(&svn->kind, svn->rev_root, svn->path, svn->p);

        if (svn->kind == svn_node_none) {
            ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
                "Invalid Repository Path: %s%s", svn->reponame, svn->path);
            return HTTP_NOT_FOUND;
        }

        if (action == NULL) {
            /*
             * Determine if svn->path is a file or a dir 
             */
            if ((svn_rev = apr_table_get(get_table, "ci"))) {
                svn->request_rev = apr_atoi64(svn_rev);
                sv_repo_setrev(r, svn);
                action = "view_commit";
            } else if (svn->kind == svn_node_dir) {
                action = "view_dir";
            }
            else {
                action = "view_inline";
            }
        }

        err =
            svn_fs_revision_root(&rev_root, svn->repo_fs, svn->last_rev,
            svn->p);
        SVN_ERROR(err, "Unable to get Revision Root %s")

            err = svn_fs_node_proplist(&prop_list, rev_root, "/", svn->p);
        SVN_ERROR(err, "Unable to get Property List: %s")

            value = apr_hash_get(prop_list, SV_PROP_XSLT, APR_HASH_KEY_STRING);
        if (value) {
            svn->xslt = apr_pstrdup(svn->p, value->data);
        }
        else {
            svn->xslt = dc->xsltpath;
        }

/*
XXXXXX: This is broken now with the new themes system
        value = apr_hash_get(prop_list, SV_PROP_CSS, APR_HASH_KEY_STRING);
        if (value) {
            svn->css = apr_pstrdup(svn->p, value->data);
        }
        else {
            svn->css = dc->csspath;
        }
*/

    }

    action_handler = sv_action_get(action);

    if (action_handler->states & SV_FLAG_ACTION_NOHEAD) {
        ret = sv_action_call(action_handler, svn, get_table);
    }
    else {
        /* The rest of our actions want a XML header */
        sv_xml_header(get_table, svn, action);

        /* We always list all the possible repos. this is for the cool link at the top */
        /*
         * TODO find another way of doing this 
         */
        ret = sv_list_repos(svn, get_table, (void *)list_repo_detail);

        if (OK != ret) {
            return ret;
        }
        if (!(action_handler->states & SV_FLAG_ACTION_NORUN)) {
            sv_action_call(action_handler, svn, get_table);
        }

        sv_xml_footer(r);
    }
    ap_rflush(r);

    return ret;
}

static void *svn_view_create_perdir_config(apr_pool_t * p, char *dir)
{
    sv_dir_config_rec *dConfig = apr_pcalloc(p, sizeof(*dConfig));

    dConfig->sv_root = apr_pstrdup(p, dir);
    dConfig->repos = apr_hash_make(p);
    dConfig->theme_dir = ap_server_root_relative(p, "share/svn-view/themes");
    dConfig->theme_name = "blueview";
    dConfig->parent_path = NULL;
    dConfig->enabled = NULL;
    return dConfig;
}

static void *svn_view_merge_perdir_config(apr_pool_t * p,
    void *base_conf, void *new_conf)
{
    return NULL;
}

static void mod_svn_view_register_hooks(apr_pool_t * p)
{
    ap_hook_handler(mod_svn_view_handler, NULL, NULL, APR_HOOK_FIRST);

    sv_action_hash = apr_hash_make(p);

    sv_action_register("list_repos", sv_list_repos, SV_FLAG_ACTION_NORUN, NULL,
        p);
    sv_action_register("view_dir", sv_view_dir, 0, NULL, p);
    sv_action_register("view_file", sv_view_file, SV_FLAG_ACTION_NOHEAD, NULL,
        p);
    sv_action_register("view_history", sv_view_history, 0, NULL, p);
    sv_action_register("view_inline", sv_view_inline, 0, NULL, p);
    sv_action_register("view_commit", sv_view_commit, 0, NULL, p);
    sv_action_register("view_blame", sv_view_blame, 0, NULL, p);
    sv_action_register("view_rss", sv_view_rss, SV_FLAG_ACTION_NOHEAD, NULL, p);
    sv_action_register("view_diff", sv_view_diff, 0, NULL, p);
    sv_action_register("view_udiff", sv_view_udiff, SV_FLAG_ACTION_NOHEAD,
                       NULL, p);
}

static const char *svn_view_conf_repo_add(cmd_parms * parms, void *conf,
    const char *name, const char *path)
{
    sv_dir_config_rec *dConfig = (sv_dir_config_rec *) conf;

    /* TODO: insert rant on why this is evil to just blindly do this. */
    /* TODO: should we validate that we can use this repository here? */
    /* TODO: Just Open and close the repo to see if it is safe. Simple Enough */
    apr_hash_set(dConfig->repos, name, APR_HASH_KEY_STRING,
        apr_pstrdup(parms->pool, path));
    return NULL;
}

static const char *svn_view_conf_parent_path(cmd_parms * parms, void *conf,
                                             const char *path)
{
    sv_dir_config_rec *dConfig = (sv_dir_config_rec *) conf;
    dConfig->parent_path = path;
    return DECLINE_CMD;
}

static const command_rec svn_view_cmds[] = {
    AP_INIT_TAKE2("SVNViewRepository", svn_view_conf_repo_add,
        NULL,
        ACCESS_CONF,
        "The path to the subversion Repository"),
    AP_INIT_TAKE1("SVNViewAuthzSvnFile", ap_set_string_slot,
        (void *)APR_OFFSETOF(sv_dir_config_rec, authzpath),
        ACCESS_CONF,
        "The absolute path to the XSLT File"),
    AP_INIT_TAKE1("SVNView", ap_set_string_slot,
        (void *)APR_OFFSETOF(sv_dir_config_rec, enabled),
        ACCESS_CONF,
        "Enable mod_svn_view for this area."),
    AP_INIT_TAKE1("SVNViewDefaultXSLT", ap_set_string_slot,
        (void *)APR_OFFSETOF(sv_dir_config_rec, xsltpath),
        ACCESS_CONF,
        "The absolute path to the XSLT File"),
    AP_INIT_TAKE1("SVNViewTheme", ap_set_string_slot,
        (void *)APR_OFFSETOF(sv_dir_config_rec, theme_name),
        ACCESS_CONF,
        "The browser Path for the CSS file"),
    AP_INIT_TAKE1("SVNViewThemesDir", ap_set_string_slot,
        (void *)APR_OFFSETOF(sv_dir_config_rec, theme_dir),
        ACCESS_CONF,
        "The browser Path for the CSS file"),
    AP_INIT_TAKE1("SVNParentPath",  svn_view_conf_parent_path, 
                  NULL,
                  ACCESS_CONF,
                  "specifies the location in the filesystem whose "  
                  "subdirectories are assumed to be Subversion repositories."),
    {NULL}
};

module AP_MODULE_DECLARE_DATA svn_view_module = {
    STANDARD20_MODULE_STUFF,
    svn_view_create_perdir_config,
    NULL,
    NULL,
    NULL,
    svn_view_cmds,
    mod_svn_view_register_hooks,
};

static void sv_xml_header(apr_table_t * get_table,
    svn_repo_rec * svn, const char *action)
{
    const char *accept_types;
    int show_xml = 0;
    request_rec *r = svn->r;
    sv_dir_config_rec *dc = ap_get_module_config(r->per_dir_config,
        &svn_view_module);

    /* Adding show_xml=1 to the Query String will turn off the XSLT Filter */
    if (NULL == apr_table_get(get_table, "show_xml")) {
        /* we do this evil hack so we send the proper content-types */
        /* TODO: Before 1.0 this should be replaced with a more sensible method */
        accept_types = apr_table_get(r->headers_in, "Accept");
        if(( accept_types != NULL) &&
           ( ap_strstr_c(accept_types, SV_XHTML_CONTENT_TYPE) != NULL ) ) {
            ap_set_content_type(r, SV_XHTML_CONTENT_TYPE);           
        }
        else {
            ap_set_content_type(r, SV_COMPAT_CONTENT_TYPE);
        }
        /* Add our XSLT Filter to make pretty HTML */
        ap_add_output_filter(SV_XSLT_FILTER_NAME, NULL, r, r->connection);
    }
    else {
        /**
         * TODO: For -release versions do we want to disable the show_xml hack? 
         *       I don't see much harm in leaving it in. Final Answer: TBD
         */
        ap_set_content_type(r, SV_XML_CONTENT_TYPE);
        show_xml = 1;
    }

    ap_rputs("<?xml version=\"1.0\"?>\n", r);

    if(show_xml == 0) {
        ap_rprintf(r, "<?xml-stylesheet type=\"text/xsl\" href=\"file://%s\"?>\n",
            svn->xslt);
    }
    ap_rprintf(r, "<sv version=\"%s\" action=\"%s\"", SVN_VIEW_VERSION,
        action);

    ap_rprintf(r, " basedir=\"%s\"", dc->sv_root);

    if (svn->reponame) {
        ap_rprintf(r, " repo=\"%s\"", svn->reponame);
    }
    if (svn->request_rev >= 0) {
        ap_rprintf(r, " revision=\"%" SVN_REVNUM_T_FMT "\"", svn->request_rev);
    }
    ap_rputs(">\n", r);
}

static void sv_xml_footer(request_rec * r)
{
    ap_rputs("</sv>", r);
}

static void output_pretty_time(svn_repo_rec * svn, apr_pool_t* pool, 
                               const char* data)
{
    apr_time_t timet;
    apr_time_exp_t xt;
    char dstr[30];
    apr_size_t len;

    svn_time_from_cstring(&timet, data, pool);
    apr_time_exp_lt(&xt, timet);
    apr_strftime(dstr, &len, sizeof(dstr), "%d/%b/%Y %H:%M", &xt);
    ap_rputs(dstr, svn->r);
}

static void sv_fill_query_string_table(request_rec * r, apr_table_t * parms)
{
    char *key;
    char *value;
    char *query_string;
    char *strtok_state;

    if (r->args == NULL) {
        return;
    }

    query_string = apr_pstrdup(r->pool, r->args);

    key = apr_strtok(query_string, "&", &strtok_state);
    while (key) {
        value = strchr(key, '=');
        if (value) {
            *value = '\0';      /* Split the string in two */
            value++;            /* Skip passed the = */
        }
        else {
            value = "1";
        }
        ap_unescape_url(key);
        ap_unescape_url(value);
        apr_table_set(parms, key, value);
/*
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
            "Found query arg: %s = %s", key, value);
*/
        key = apr_strtok(NULL, "&", &strtok_state);
    }
}

static char *sv_build_query_string(apr_pool_t * p, apr_table_t * args)
{
    char qs[MAX_STRING_LEN];
    const apr_array_header_t *tarry;
    const apr_table_entry_t *telts;
    int i;

    memset(qs, 0, MAX_STRING_LEN);
    tarry = apr_table_elts(args);
    telts = (apr_table_entry_t *) tarry->elts;
    for (i = 0; i < tarry->nelts; i++) {
        if (i > 0) {
            strncpy(qs + strlen(qs), "&amp;", (MAX_STRING_LEN - strlen(qs)));
        }
        apr_snprintf(qs + strlen(qs), (MAX_STRING_LEN - strlen(qs)),
            "%s=%s", telts[i].key, telts[i].val);
    }
    return apr_pstrdup(p, qs);
}

static int sv_view_dir(svn_repo_rec * svn, apr_table_t * parms, void *data)
{
    request_rec *r = svn->r;
    svn_error_t *err;
    apr_hash_t *dirents;
    apr_hash_t *prop_list;
    apr_hash_index_t *dirwalk;
    svn_string_t *propvalue;
    const char *dirpath = apr_pstrndup(r->pool, svn->path,
        strlen(svn->path));
    sv_dir_config_rec *dc;

    dc = ap_get_module_config(r->per_dir_config, &svn_view_module);

    /*
     * trim off trailing slash except for root directory 
     */
    if (strlen(dirpath) == 0) {
        dirpath = "/";
    }

    ap_rprintf(r, "<dirlist name=\"%s\">\n", dirpath);

    sv_repo_getprops(r, svn, svn->path, &prop_list);
    sv_show_props(r, svn, prop_list);

    err = svn_fs_dir_entries(&dirents, svn->rev_root, svn->path, svn->p);
    if (err != SVN_NO_ERROR) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, err->apr_err, r,
            "Directory list returned: %s", err->message);
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    for (dirwalk = apr_hash_first(svn->p, dirents);
        dirwalk != NULL; dirwalk = apr_hash_next(dirwalk)) {
        char *key;
        int klen;
        svn_fs_dirent_t *val;
        char *full_path;
        svn_revnum_t lastrev;

        apr_hash_this(dirwalk, (void *)&key, &klen, (void **)&val);
        apr_filepath_merge(&full_path, svn->path, val->name,
            APR_FILEPATH_TRUENAME, svn->p);
        ap_rprintf(r, "<node name=\"%s\" path=\"%s\"", val->name, svn->path);

        svn_fs_node_created_rev(&lastrev, svn->rev_root, full_path, svn->p);

        if (val->kind == svn_node_dir) {
            ap_rputs(" type=\"dir\"", r);
        }
        else {
            ap_rputs(" type=\"file\"", r);
        }
        ap_rprintf(r, " revision=\"%" SVN_REVNUM_T_FMT "\">\n", lastrev);
        /*
         * Pull properties for file 
         */
        svn_fs_revision_prop(&propvalue, svn->repo_fs, lastrev,
            SVN_PROP_REVISION_AUTHOR, svn->p);
        if (propvalue) {
            ap_rprintf(r, "<author>%s</author>\n",
                ap_escape_html(svn->p, propvalue->data));
        }
        svn_fs_revision_prop(&propvalue, svn->repo_fs, lastrev,
            SVN_PROP_REVISION_LOG, svn->p);
        if (propvalue) {
            ap_rprintf(r, "<log>%s</log>\n",
                ap_escape_html(svn->p, propvalue->data));
        }
        svn_fs_revision_prop(&propvalue, svn->repo_fs, lastrev,
            SVN_PROP_REVISION_DATE, svn->p);
        if (propvalue) {
            ap_rputs("<date>", svn->r);
            output_pretty_time(svn, svn->p, propvalue->data);
            ap_rputs("</date>", svn->r);
        }
        svn_fs_revision_prop(&propvalue, svn->repo_fs, lastrev,
            SVN_PROP_REVISION_ORIG_DATE, svn->p);
        if (propvalue) {
            ap_rprintf(r, "<dateorig>%s</origdate>\n", propvalue->data);
        }
        ap_rputs("</node>\n", r);
    }
    ap_rputs("</dirlist>\n", r);

    return OK;
}

/**
 *  TODO: 
 *     - Add Line Numbers
 *     - Add Links to different actions.
 */
static int sv_view_inline(svn_repo_rec * svn, apr_table_t * parms, void *data)
{
    request_rec *r = svn->r;
    svn_string_t *mime_type;
    apr_hash_t *prop_list;
    svn_revnum_t lastrev;
    svn_string_t *propvalue;
    request_rec *rr = NULL;
    ap_filter_t *f;
    sv_dir_config_rec *dc = ap_get_module_config(r->per_dir_config,
        &svn_view_module);

    if (svn->path == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, svn->r,
            "'path' is a required parameter for this Action.");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    ap_rprintf(r, "<file name=\"%s\"", svn->path);

    svn_fs_node_created_rev(&lastrev, svn->rev_root, svn->path, svn->p);

    ap_rprintf(r, " revision=\"%" SVN_REVNUM_T_FMT "\">\n", lastrev);
    /*
     * Pull properties for file 
     */
    svn_fs_revision_prop(&propvalue, svn->repo_fs, lastrev,
        SVN_PROP_REVISION_AUTHOR, svn->p);
    if (propvalue) {
        ap_rprintf(r, "<author>%s</author>\n",
            ap_escape_html(svn->p, propvalue->data));
    }
    svn_fs_revision_prop(&propvalue, svn->repo_fs, lastrev,
        SVN_PROP_REVISION_LOG, svn->p);
    if (propvalue) {
        ap_rprintf(r, "<log>%s</log>\n",
            ap_escape_html(svn->p, propvalue->data));
    }
    svn_fs_revision_prop(&propvalue, svn->repo_fs, lastrev,
        SVN_PROP_REVISION_DATE, svn->p);
    if (propvalue) {
        ap_rputs("<date>", svn->r);
        output_pretty_time(svn, svn->p, propvalue->data);
        ap_rputs("</date>", svn->r);
    }
    svn_fs_revision_prop(&propvalue, svn->repo_fs, lastrev,
        SVN_PROP_REVISION_ORIG_DATE, svn->p);
    if (propvalue) {
        ap_rprintf(r, "<dateorig>%s</origdate>\n", propvalue->data);
    }

    sv_repo_getprops(r, svn, svn->path, &prop_list);
    sv_show_props(r, svn, prop_list);

    mime_type = apr_hash_get(prop_list, "svn:mime-type", APR_HASH_KEY_STRING);

    if (mime_type == NULL || sv_viewable_content_type(mime_type) == 0) {
        if(mime_type == NULL) {
            ap_rputs("<contents mimetype=\"text/plain\" type=\"text\">\n<![CDATA[\n",r);
        } 
        else {
            ap_rprintf(r, "<contents mimetype=\"%s\" type=\"text\">\n",
                mime_type->data);
        }

        rr = ap_sub_req_lookup_uri(apr_psprintf(svn->p, "%s/%s%s?action=view_file&colorer=1&colorer_html_wrappers=0", dc->sv_root, svn->reponame, svn->path), r, r->output_filters);

        for (f=rr->output_filters;
            f->frec != ap_subreq_core_filter_handle; f = f->next) {
            f->next = r->output_filters;
        }

        ap_run_sub_req(rr);

        if(mime_type == NULL) {
            ap_rputs("\n]]>", r);
        }
        ap_rputs("\n</contents>\n", r);
    }
    else if (sv_viewable_content_type(mime_type) == 1) {
        ap_rprintf(r, "<contents mimetype=\"%s\" type=\"image\">\n",
            mime_type->data);
        ap_rputs("\n</contents>\n", r);
    }
    else {
        ap_rprintf(r, "<contents mimetype=\"%s\" type=\"binary\">\n",
            mime_type->data);
        ap_rputs("\n</contents>\n", r);
    }
    ap_rputs("\n</file>\n", r);
    return OK;
}

/**
 * The Following two functions where basicly copied from 
 * from subversion/svnlook/main.c 
 */
/* Generate a generic delta tree. */
APR_DECLARE( svn_error_t *) sv_generate_delta_tree(svn_repos_node_t ** tree,
    svn_repo_rec * svn,
    svn_revnum_t base_rev, svn_boolean_t use_copy_history)
{
    svn_fs_root_t *base_root;
    svn_fs_root_t *rev_root;
    const svn_delta_editor_t *editor;
    void *edit_baton;
    apr_pool_t *edit_pool;
    apr_pool_create(&edit_pool, svn->p);

    /*
     * Get the base root. 
     */
    svn_fs_revision_root(&base_root, svn->repo_fs, base_rev-1, svn->p);
    svn_fs_revision_root(&rev_root, svn->repo_fs, base_rev, svn->p);
    /*
     * Request our editor. 
     */
    svn_repos_node_editor(&editor, &edit_baton, svn->repo,
        base_root, rev_root, svn->p, edit_pool);

    /*
     * Drive our editor. 
     */
    svn_repos_replay(rev_root, editor, edit_baton, edit_pool);

    /*
     * Return the tree we just built. 
     */
    *tree = svn_repos_node_from_baton(edit_baton);
    apr_pool_destroy(edit_pool);
    return SVN_NO_ERROR;
}

/* Recursively print all nodes in the tree that have been modified
   (do not include directories affected only by "bubble-up"). */
static svn_error_t *sv_print_changed_tree(svn_repos_node_t * node,
    const char *path /* UTF-8! */ ,
    svn_repo_rec* svn)
{
    const char *full_path;
    char status[3] = "_ ";
    int print_me = 1;

    if (!node)
        return SVN_NO_ERROR;

    /* Check the Node's Authz Permissions */
    full_path = svn_path_join(path, node->name, svn->p);
    if (!sv_authz_check(svn, full_path)) {
        return SVN_NO_ERROR;
    }

    /*
     * Print the node. 
     */
    if (node->action == 'A')
        status[0] = 'A';
    else if (node->action == 'D')
        status[0] = 'D';
    else if (node->action == 'R') {
        if ((!node->text_mod) && (!node->prop_mod)) {
            print_me = 0;
        }
        if (node->text_mod) {
            status[0] = 'U';
        }
        if (node->prop_mod) {
            status[1] = 'U';
        }
    }
    else {
        print_me = 0;
    }

    /*
     * Print this node unless told to skip it. 
     */
    if (print_me) {
        const char *path_stdout;

        svn_utf_cstring_from_utf8(&path_stdout, path, svn->p);
        ap_rprintf(svn->r, "<modfile action=\"%s\" path=\"%s%s\"/>\n", 
            status, path_stdout, node->kind == svn_node_dir ? "/" : "");
    }

    /*
     * Return here if the node has no children. 
     */
    node = node->child;
    if (!node)
        return SVN_NO_ERROR;

    /*
     * Recursively handle the node's children. 
     */
    sv_print_changed_tree(node, full_path, svn);
    while (node->sibling) {
        node = node->sibling;
        full_path = svn_path_join(path, node->name, svn->p);
        sv_print_changed_tree(node, full_path, svn);
    }

    return SVN_NO_ERROR;
}


/* TODO: sv_show_revision: Note Path Changes so we can do trunk/branch/tag magic */
/* TODO: sv_show_revision: Make a list of all changed files by this commit */
static svn_error_t *sv_show_revision(void *p_baton, const char
    *path, svn_revnum_t revision, apr_pool_t * pool)
{
    svn_repos_node_t *tree;
    svn_string_t *propvalue;
    apr_pool_t *subpool;
    svn_repo_rec *svn;
    sv_history_baton_st *baton = (sv_history_baton_st*) p_baton;

    svn = baton->svn;

    apr_pool_create(&subpool, pool);

    ap_rprintf(svn->r, "<revision id=\"%" SVN_REVNUM_T_FMT "\">\n", revision);

    svn_fs_revision_prop(&propvalue, svn->repo_fs, revision,
        SVN_PROP_REVISION_AUTHOR, subpool);
    if (propvalue) {
        ap_rprintf(svn->r, "<author>%s</author>\n",
            ap_escape_html(subpool, propvalue->data));
    }

    svn_fs_revision_prop(&propvalue, svn->repo_fs, revision,
        SVN_PROP_REVISION_LOG, subpool);
    if (propvalue) {
        ap_rprintf(svn->r, "<log>%s</log>\n",
            ap_escape_html(subpool, propvalue->data));
    }

    svn_fs_revision_prop(&propvalue, svn->repo_fs, revision,
        SVN_PROP_REVISION_DATE, subpool);
    if (propvalue) {
        ap_rputs("<date>", svn->r);
        output_pretty_time(svn, subpool, propvalue->data);
        ap_rputs("</date>", svn->r);
    }

    svn_fs_revision_prop(&propvalue, svn->repo_fs, revision,
        SVN_PROP_REVISION_ORIG_DATE, subpool);
    if (propvalue) {
        ap_rprintf(svn->r, "<dateorig>%s</origdate>\n", propvalue->data);
    }

    if (baton->show_files == 1) {
        sv_generate_delta_tree(&tree, svn, revision,
            TRUE);
        ap_rputs("<modfiles>\n", svn->r);
        sv_print_changed_tree(tree,"/",svn);
        ap_rputs("</modfiles>\n", svn->r);
    }
   
    ap_rputs("</revision>\n", svn->r);

    apr_pool_destroy(subpool);

    if(1 > --baton->num_left) {
        return svn_error_create(-1, NULL, "");
    }
    else {
        return SVN_NO_ERROR;
    }
}
static int sv_view_history(svn_repo_rec * svn, apr_table_t * parms, void *data)
{
    /**
     * TODO: When view_history is ran on a directory, 
     *        the directory links at the top of a page are Wrong. 
     *        Should fix this before we release.
     */
    svn_revnum_t lastrev;
    sv_history_baton_st *baton;
    char * d1str;
    request_rec *r = svn->r;

    if (svn->path == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, svn->r,
            "'path' is a required parameter for this Action.");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    ap_rprintf(r, "<file name=\"%s\">", svn->path);
    ap_rputs("<history>\n", r);

    d1str = apr_table_get(parms, "d1");

    if(d1str) {
        int d1rev = atoi(d1str);
        if(d1rev > 0) {
            ap_rprintf(r, "<diff d1=\"%d\"/>", d1rev);
        }
    }

    svn_fs_node_created_rev(&lastrev, svn->rev_root, svn->path, svn->p);

    baton = (sv_history_baton_st *)apr_palloc(svn->p, sizeof(*baton));
    baton->num_left = 100;
    baton->svn = svn;
    baton->show_files = 0;

    svn_repos_history2(svn->repo_fs, svn->path, 
        sv_show_revision, baton, 
        sv_authz_callback, svn,
        0, lastrev, 1, svn->p);

    ap_rputs("\n</history>\n", r);
    ap_rputs("\n</file>\n", r);
    return OK;
}

static int sv_view_file(svn_repo_rec * svn, apr_table_t * parms, void *data)
{
    request_rec *r = svn->r;
    char *block;
    svn_string_t *mime_type;
    svn_error_t *err;
    svn_stream_t *contents;
    apr_hash_t *prop_list;
    const char* colorer;

    if (svn->path == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, svn->r,
            "'path' is a required parameter for this Action.");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    err = svn_fs_file_contents(&contents, svn->rev_root, svn->path, svn->p);

    if (err != SVN_NO_ERROR) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, err->apr_err, r,
            "SVN File Contents Returned: %s", err->message);
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    sv_repo_getprops(r, svn, svn->path, &prop_list);

    mime_type = apr_hash_get(prop_list, "svn:mime-type", APR_HASH_KEY_STRING);
    colorer = apr_table_get(parms, "colorer");
    if (mime_type == NULL) {
        /* Is this a good default type? */
        ap_set_content_type(r, "text/plain");
    }
    else {
        ap_set_content_type(r,
            apr_pstrmemdup(svn->p, mime_type->data, mime_type->len));
        if(sv_viewable_content_type(mime_type) != 0) {
            colorer = NULL;
        }
    }

    if(colorer != NULL) {
        ap_add_output_filter(SV_HIGHLIGHT_FILTER_NAME, NULL, r, r->connection); 
    }        
    
    /* TODO: Also send out headers so browsers see 'nice' filenames. */

    block = apr_palloc(svn->p, SVN_STREAM_CHUNK_SIZE);

    while (1) {
        apr_size_t bufsize = SVN_STREAM_CHUNK_SIZE;

        err = svn_stream_read(contents, block, &bufsize);
        if (err != SVN_NO_ERROR) {
            ap_rprintf(r, "Error Reading File Stream: %s", err->message);
            ap_log_rerror(APLOG_MARK, APLOG_ERR, err->apr_err, r,
                "SVN Stream Read Returned: %s", err->message);
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        /* End of File. */
        if (bufsize == 0) {
            break;
        }

        if(ap_rwrite(block, bufsize, r) <= 0) {
            break;
        }
    }
    svn_stream_close(contents);

    ap_rflush(r);

    return OK;
}

static int sv_view_commit(svn_repo_rec * svn, apr_table_t * parms, void *data)
{
    request_rec *r = svn->r;
    sv_history_baton_st *baton;

    ap_rprintf(r, "<commit rev=\"%" SVN_REVNUM_T_FMT "\">", svn->request_rev);

    baton = (sv_history_baton_st *)apr_palloc(svn->p, sizeof(*baton));

    baton->num_left = 1;
    baton->svn = svn;
    baton->show_files = 1;

    svn_repos_history2(svn->repo_fs, svn->path, 
        sv_show_revision, baton, 
        sv_authz_callback, svn,
        0, svn->request_rev, 1, svn->p);

    ap_rputs("</commit>\n", r);
    
    return OK;
}


static int sv_list_repos(svn_repo_rec * svn, apr_table_t * parms, void *data)
{
    request_rec *r = svn->r;
    apr_hash_index_t *hi;
    sv_dir_config_rec *dc = ap_get_module_config(r->per_dir_config,
        &svn_view_module);

    ap_rputs("<repos>\n", r);
    for (hi = apr_hash_first(svn->p, svn->repos); hi != NULL;
        hi = apr_hash_next(hi)) {
        char *key;
        int klen;
        char *val;

        apr_hash_this(hi, (void *)&key, &klen, (void **)&val);
        if (0 == data) {
            ap_rprintf(r, "<repo short_name=\"%s\"/>\n", key);
        }
        else {
            svn_error_t *err;
            svn_repos_t *repo;
            svn_fs_t *repo_fs;
            svn_fs_root_t *rev_root;
            svn_revnum_t last_rev;
            svn_string_t *value;
            apr_hash_t *prop_list;
            const char *path = "/";

            ap_rprintf(r, "<repo short_name=\"%s\">\n", key);

            /*
             * TODO: We should try to cache this whole section, 
             *       maybe fetch values on-conf creation?
             */
            
            err = svn_repos_open(&repo, val, svn->p);
            SVN_ERROR(err, "Unable to Open Repos: %s")

                repo_fs = svn_repos_fs(repo);

            err = svn_fs_youngest_rev(&last_rev, repo_fs, svn->p);
            SVN_ERROR(err, "Unable to find the Youngest Revision: %s")

                err = svn_fs_revision_root(&rev_root, repo_fs,
                last_rev, svn->p);
            SVN_ERROR(err, "Unable to find the Revision Root: %s")

                err = svn_fs_node_proplist(&prop_list, rev_root, path, svn->p);
            SVN_ERROR(err, "Unable to get Property List: %s")

                value =
                apr_hash_get(prop_list, SV_PROP_DESC, APR_HASH_KEY_STRING);

            if (value) {
                ap_rprintf(r, "<desc>%s</desc>\n",
                    ap_escape_html(svn->p, value->data));
            }

            /* TODO: Other Properties that we need to show?? */
            ap_rputs("</repo>\n", r);
        }
    }
    ap_rputs("</repos>\n", r);

    return OK;
}

static int sv_show_props(request_rec * r, svn_repo_rec * svn,
    apr_hash_t * prop_list)
{
    apr_hash_index_t *prop_walk;

    ap_rprintf(r, "<properties count='%d'>\n", apr_hash_count(prop_list));
    for (prop_walk = apr_hash_first(svn->p, prop_list);
        prop_walk != NULL; prop_walk = apr_hash_next(prop_walk)) {
        char *key;
        int klen;
        svn_string_t *val;

        apr_hash_this(prop_walk, (void *)&key, &klen, (void **)&val);
        ap_rputs("<prop name=\"", r);
        ap_rputs(ap_escape_html(svn->p, key), r);
        ap_rputs("\" value=\"", r);
        ap_rputs(ap_escape_html(svn->p, val->data), r);
        ap_rputs("\"/>\n", r);
    }
    ap_rputs("</properties>\n", r);
    return OK;
}

/**
 * return values:
 *  0 == text 
 *  1 == image 
 *  2 == unknown 
 */
static int sv_viewable_content_type(svn_string_t * mime_type)
{
    if (mime_type == NULL)
        return 2;

    if (mime_type->len > strlen("image/")) {
        if (strncmp(mime_type->data, "image/", strlen("image/")) == 0) {
            return 1;
        }
    }

    if (mime_type->len > strlen("text/")) {
        if (strncmp(mime_type->data, "text/", strlen("text/")) == 0) {
            return 0;
        }
    }

    return 2;
}

static int sv_repo_open(request_rec * r, svn_repo_rec * svn)
{
    svn_error_t *err;

    err = svn_repos_open(&svn->repo, svn->repopath, svn->p);
    SVN_ERROR(err, "Unable to Open Repos: %s")

        svn->repo_fs = svn_repos_fs(svn->repo);

    err = svn_fs_youngest_rev(&svn->last_rev, svn->repo_fs, svn->p);

    SVN_ERROR(err, "Unable to find the Youngest Revision: %s")

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
        "sv_open_repo returned OK");
    return APR_SUCCESS;
}

static int sv_repo_setrev(request_rec * r, svn_repo_rec * svn)
{
    svn_error_t *err;

    if (!svn->repo)
        return -1;
    if (svn->request_rev == SVN_INVALID_REVNUM) {
        svn->request_rev = svn->last_rev;
    }
    err =
        svn_fs_revision_root(&svn->rev_root, svn->repo_fs, svn->request_rev,
        svn->p);
    SVN_ERROR(err, "Unable to find the Revision Root: %s")
        return OK;
}

/**
 * Get visible properties
 *
 */
static int sv_repo_getprops(request_rec * r, svn_repo_rec * svn,
    const char *svn_path, apr_hash_t ** prop_list)
{
    svn_error_t *err;

    if (!svn->repo)
        return -1;

    err = svn_fs_node_proplist(&*prop_list, svn->rev_root, svn_path, svn->p);
    SVN_ERROR(err, "Unable to get Property List: %s")
        return APR_SUCCESS;
}


APR_DECLARE(void)sv_action_register(const char *action,
    sv_action_handler * func, int states, void *data, apr_pool_t * p)
{
    char *dupkey = apr_pstrdup(p, action);
    sv_action_handler_st *hand = apr_palloc(p, sizeof(*hand));

    hand->func = func;
    hand->states = states;
    hand->data = data;
    ap_str_tolower(dupkey);
    apr_hash_set(sv_action_hash, dupkey, APR_HASH_KEY_STRING, hand);
}
