/* ====================================================================
 *  Copyright 2004 Edward Rudd and Paul Querna
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Some code based off of mod_dav_svn:
 *     Copyright (c) CollabNet.  All rights reserved.
 *
 */

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "apr_tables.h"
#include "apr_hash.h"
#include "svn_repos.h"
#include "svn_io.h"
#include "svn_path.h"
#include "svn_diff.h"
#include "mod_svn_view_private.h"
#include "svn_stream_apr_brigade.h"

#define SVN_ERROR_X(file, line, err,msg)  \
   if (err != SVN_NO_ERROR) { \
        char* m = apr_pstrcat(r->pool, "[%s:%d] ", msg, NULL); \
        ap_log_rerror(APLOG_MARK, APLOG_ERR, err->apr_err, r, \
            m, file, line, err->message); \
        return HTTP_INTERNAL_SERVER_ERROR; \
    }
#define SVN_ERROR(err,msg)  SVN_ERROR_X(__FILE__, __LINE__, err, msg)

#define SVN_ERROR_APR_X(file, line, err, msg)  \
   if (err != SVN_NO_ERROR) { \
        char* m = apr_pstrcat(svn->p, "[%s:%d] ", msg, NULL); \
        ap_log_rerror(APLOG_MARK, APLOG_ERR, err->apr_err, svn->r, \
            m, file, line, err->message); \
        return APR_EGENERAL; \
    }
#define SVN_ERROR_APR(err,msg)  SVN_ERROR_APR_X(__FILE__, __LINE__, err, msg);

#define SV_HIGHLIGHT_FILTER_NAME "HIGHLIGHT"


int sv_view_diff(svn_repo_rec * svn, apr_table_t * parms, void *data);
int sv_view_udiff(svn_repo_rec * svn, apr_table_t * parms, void *data);
static apr_status_t sv_diff_rev_to_tmpfile(svn_repo_rec * svn, apr_pool_t * p,
                                           const char *temp_dir, int rev,
                                           const char **tmp_file);
static apr_status_t sv_diff_create_udiff(svn_repo_rec * svn, int rev1,
                                         int rev2);

static int sv_string_to_revision(svn_repo_rec * svn, const char* rev);


int sv_view_diff(svn_repo_rec * svn, apr_table_t * parms, void *data)
{
    return OK;
}

int sv_view_udiff(svn_repo_rec * svn, apr_table_t * parms, void *data)
{
    int r1 = 0;
    int r2 = 0;
    const char* colorer = apr_table_get(parms, "colorer");
    const char *rev = apr_table_get(parms, "r1");
    if (!rev) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, svn->r,
                      "r1 of udiff was invalid");
        return HTTP_INTERNAL_SERVER_ERROR;
    }
    r1 = sv_string_to_revision(svn, rev);

    rev = apr_table_get(parms, "r2");
    if (!rev) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, svn->r,
                      "r2 of udiff was invalid");
        return HTTP_INTERNAL_SERVER_ERROR;
    }
    r2 = sv_string_to_revision(svn, rev);

    if (r2 == SVN_INVALID_REVNUM || r1 == SVN_INVALID_REVNUM) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, svn->r,
                      "r1 or r2 are invalid");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    if(colorer != NULL) {
        ap_add_output_filter(SV_HIGHLIGHT_FILTER_NAME, NULL, svn->r, svn->r->connection);
    }

    if (sv_diff_create_udiff(svn, r1, r2) != APR_SUCCESS) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, svn->r,
                      "create_udiff() failed!");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    return OK;
}

static int sv_string_to_revision(svn_repo_rec * svn, const char* rev) 
{
    if(0 ==  apr_strnatcasecmp("HEAD",rev) ||  0 ==  apr_strnatcasecmp("TRUNK", rev)) {
        return svn->last_rev;
    } 
    else {
        int r = apr_atoi64(rev);
        if(r <= 0) {
            return SVN_INVALID_REVNUM;
        }
        else {
            return r;
        }
    }
}

/**
 *  TODO: Error`ing out of this function might leave things open/behind. 
 *        We need better cleanup! 
 */
static apr_status_t sv_diff_create_udiff(svn_repo_rec * svn, int rev1,
                                         int rev2)
{
    const char *fn_r1;
    const char *fn_r2;
    const char *temp_dir;
    apr_pool_t *p;
    apr_status_t ret;
    svn_diff_t *diff;
    svn_stream_t *udiff_stream;
    apr_bucket_brigade *bb;

    apr_pool_create(&p, svn->p);

    /* Diff Header */
    ap_rprintf(svn->r, "Index: %s"APR_EOL_STR, svn->path);
    ap_rputs("==================================================================="APR_EOL_STR,
             svn->r);

    bb = apr_brigade_create(svn->r->pool, svn->r->connection->bucket_alloc);
    svn_io_temp_dir(&temp_dir, p);

    ret = sv_diff_rev_to_tmpfile(svn, p, temp_dir, rev1, &fn_r1);

    if (ret != APR_SUCCESS) {
        return ret;
    }


    ret = sv_diff_rev_to_tmpfile(svn, p, temp_dir, rev2, &fn_r2);

    if (ret != APR_SUCCESS) {
        apr_file_remove(fn_r1, p);
        apr_pool_destroy(p);
        return ret;
    }

    svn_diff_file_diff(&diff, fn_r1, fn_r2, p);

    udiff_stream = svn_stream_from_aprbrigade(bb, svn->r->pool);

    svn_diff_file_output_unified(udiff_stream, diff, fn_r1, fn_r2, 
                                 apr_psprintf(p, "%s (revision %d)", svn->path, rev1), 
                                 apr_psprintf(p, "%s (revision %d)", svn->path, rev2), 
                                 p);

    ap_pass_brigade(svn->r->output_filters, bb);

    apr_file_remove(fn_r1, p);
    apr_file_remove(fn_r2, p);

    apr_pool_destroy(p);

    return APR_SUCCESS;
}

/**
 *  TODO: Error`ing out of this function might leave things open/behind. 
 *        We need better cleanup! 
 */
/**
 *  TODO: Who checks in a revision if valid?  This function sure doesn't.
 */
static apr_status_t sv_diff_rev_to_tmpfile(svn_repo_rec * svn, apr_pool_t * p,
                                           const char *temp_dir, int rev,
                                           const char **tmp_file)
{
    apr_file_t *file;
    const char *filename;
    svn_stream_t *fstream;
    char *block;
    svn_error_t *err;
    svn_stream_t *contents;
    svn_fs_root_t *rev_root;

    err = svn_fs_revision_root(&rev_root, svn->repo_fs, rev, p);
    SVN_ERROR_APR(err, "Unable to get revision root");

    err = svn_fs_file_contents(&contents, rev_root, svn->path, p);
    SVN_ERROR_APR(err, "Unable to get file contents");

    err = svn_io_open_unique_file(&file, &filename,
                                  svn_path_join(temp_dir, "view_diff", p),
                                  "", FALSE, p);
    SVN_ERROR_APR(err, "Unable to open unique file");

    fstream = svn_stream_from_aprfile(file, p);

    block = apr_palloc(p, SVN_STREAM_CHUNK_SIZE);

    while (1) {
        apr_size_t bufsize = SVN_STREAM_CHUNK_SIZE;

        err = svn_stream_read(contents, block, &bufsize);
        SVN_ERROR_APR(err, "Error Reading File Contents");

        /* End of File. */
        if (bufsize == 0)
            break;

        /* TODO: What if stream_write doesn't write the full block? */
        svn_stream_write(fstream, block, &bufsize);
    }
    svn_stream_close(contents);
    svn_stream_close(fstream);

    svn_io_file_close(file, p);

    *tmp_file = filename;

    return APR_SUCCESS;
}

