/* ====================================================================
 *  Copyright 2004 Edward Rudd and Paul Querna
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Some code based off of mod_dav_svn:
 *     Copyright (c) CollabNet.  All rights reserved.
 *
 */

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "apr_tables.h"
#include "apr_hash.h"
#include "apr_time.h"
#include "svn_repos.h"
#include "svn_path.h"
#include "svn_utf.h"
#include "svn_time.h"
#include "svn_props.h"
#include "mod_svn_view_private.h"

#define SV_RSS_MAX_CHANGES 100
#define SV_RSS_DEFAULT_CHANGES 10

#define SV_RSS_CONTENT_TYPE "application/xml"

static void sv_rss_header(svn_repo_rec * svn);
static void sv_rss_footer(svn_repo_rec * svn);
static const char* sv_rss_convert_date(const char* str, apr_pool_t* p);
static svn_error_t *sv_show_rss_item(void *p_baton, const char
    *path, svn_revnum_t revision, apr_pool_t * pool);
static svn_error_t *sv_rss_print_changed_tree(svn_repos_node_t * node,
    const char *path,
    svn_repo_rec* svn);


typedef struct sv_history_baton_st sv_history_baton_st;
struct sv_history_baton_st {
    svn_repo_rec *svn;
    int num_left;
};

#define SVN_ERROR(err,msg)  \
   if (err != SVN_NO_ERROR) { \
        ap_log_rerror(APLOG_MARK, APLOG_ERR, err->apr_err, r, \
            msg, err->message); \
        return HTTP_INTERNAL_SERVER_ERROR; \
    }

int sv_view_rss(svn_repo_rec * svn, apr_table_t * get_table, void *data) 
{
    svn_revnum_t lastrev;
    sv_history_baton_st *baton;
    int changes = 0;
    const char* num_changes;

    if (svn->path == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, svn->r,
            "The Path must be set for view_rss");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    num_changes = apr_table_get(get_table, "n");
    if(num_changes != NULL) {
        changes = (int)strtol(num_changes, (char **)NULL, 10);
        if(changes > SV_RSS_MAX_CHANGES || changes < 1) {
            changes = SV_RSS_DEFAULT_CHANGES;
        }
    }
    else {
        changes = SV_RSS_DEFAULT_CHANGES;
    }

    sv_rss_header(svn);

    lastrev = svn->request_rev;
    baton = (sv_history_baton_st *)apr_palloc(svn->p, sizeof(*baton));
    baton->num_left = changes;
    baton->svn = svn;

    svn_repos_history(svn->repo_fs,
        svn->path, sv_show_rss_item, (void *)baton, 0, lastrev, 1, svn->p);

    sv_rss_footer(svn);
    return OK;
}

static svn_error_t *sv_show_rss_item(void *p_baton, const char
    *path, svn_revnum_t revision, apr_pool_t * pool)
{
    const char *datestr;
    svn_repos_node_t *tree;
    svn_string_t *propvalue;
    const char *repo_url;
    svn_repo_rec *svn;
    sv_history_baton_st *baton = (sv_history_baton_st*) p_baton;
    svn = baton->svn;

    repo_url = apr_pstrcat(pool, svn->sv_root, "/", svn->reponame, "/?action=view_history&amp;rev=",  apr_psprintf(svn->p, "%" SVN_REVNUM_T_FMT, revision), NULL);
    ap_rputs("<item>\n",svn->r);

    svn_fs_revision_prop(&propvalue, svn->repo_fs, revision,
        SVN_PROP_REVISION_AUTHOR, pool);
    if (propvalue) {
        ap_rprintf(svn->r, "<author>%s</author>\n",
            ap_escape_html(pool, propvalue->data));

        ap_rprintf(svn->r, "<title>r%" SVN_REVNUM_T_FMT " by %s</title>", revision, ap_escape_html(pool, propvalue->data));
    }
    else {
        ap_rprintf(svn->r, "<title>r%" SVN_REVNUM_T_FMT "</title>", revision);
    }
    ap_rprintf(svn->r, "<link>%s</link>\n",ap_construct_url(pool, repo_url, svn->r));
    ap_rputs("<description>\n&lt;div style=\"white-space: pre;\"&gt;",svn->r);

    svn_fs_revision_prop(&propvalue, svn->repo_fs, revision,
        SVN_PROP_REVISION_LOG, pool);
    if (propvalue) {
        ap_rprintf(svn->r, "%s\n",
            ap_escape_html(pool, propvalue->data));
    }
    ap_rputs("&lt;/div&gt;&lt;br/&gt;&lt;div style=\"white-space: pre;\"&gt;Changed Files:\n",svn->r);

    sv_generate_delta_tree(&tree, svn, revision, TRUE);
    sv_rss_print_changed_tree(tree,"/",svn);
    ap_rputs("&lt;/div&gt;\n</description>\n", svn->r);

    svn_fs_revision_prop(&propvalue, svn->repo_fs, revision,
        SVN_PROP_REVISION_DATE, pool);
    if (propvalue) {
        datestr = propvalue->data;
        ap_rprintf(svn->r, "<pubDate>%s</pubDate>\n", sv_rss_convert_date(datestr,pool));
    }
    ap_rputs("</item>\n",svn->r);
    if(1 > --baton->num_left) {
        return svn_error_create(-1, NULL, "");
    }
    else {
        return SVN_NO_ERROR;
    }
}


static const char* sv_rss_convert_date(const char* str, apr_pool_t* p) {
    char date[APR_RFC822_DATE_LEN];
    apr_time_t time;
    svn_time_from_cstring (&time, str, p);
    apr_rfc822_date(date, time);
    return apr_pstrdup(p, date);
}

static void sv_rss_header(svn_repo_rec * svn)
{
    const char *repo_url = apr_pstrcat(svn->p, svn->sv_root, "/", svn->reponame, "/", NULL);

    ap_set_content_type(svn->r, SV_RSS_CONTENT_TYPE);

    ap_rputs("<?xml version=\"1.0\"?>\n", svn->r);
    ap_rputs("<rss version=\"2.0\">\n"
             "<channel>\n"
             "<generator>mod_svn_view_rss</generator>\n"
             "<description>Changes to this Subversion Repository</description>\n",svn->r);
    ap_rprintf(svn->r, "<title>Changes for %s</title>\n", svn->reponame);
    ap_rprintf(svn->r, "<link>%s</link>\n",ap_construct_url(svn->p, repo_url, svn->r));
}

static void sv_rss_footer(svn_repo_rec * svn) {
    ap_rputs("</channel>\n</rss>\n", svn->r);
}

static svn_error_t *sv_rss_print_changed_tree(svn_repos_node_t * node,
    const char *path /* UTF-8! */ ,
    svn_repo_rec* svn)
{
    const char *full_path;
    char status[3] = "_ ";
    int print_me = 1;

    if (!node)
        return SVN_NO_ERROR;

    /*
     * Print the node.
     */
    if (node->action == 'A')
        status[0] = 'A';
    else if (node->action == 'D')
        status[0] = 'D';
    else if (node->action == 'R') {
        if ((!node->text_mod) && (!node->prop_mod))
            print_me = 0;
        if (node->text_mod)
            status[0] = 'U';
        if (node->prop_mod)
            status[1] = 'U';
    }
    else
        print_me = 0;

    /*
     * Print this node unless told to skip it.
     */
    if (print_me) {
        const char *path_stdout;

        svn_utf_cstring_from_utf8(&path_stdout, path, svn->p);
        ap_rprintf(svn->r, "%s  %s%s\n",
            status, path_stdout, node->kind == svn_node_dir ? "/" : "");
    }

    /*
     * Return here if the node has no children.
     */
    node = node->child;
    if (!node)
        return SVN_NO_ERROR;

    /*
     * Recursively handle the node's children.
     */
    full_path = svn_path_join(path, node->name, svn->p);
    sv_rss_print_changed_tree(node, full_path, svn);
    while (node->sibling) {
        node = node->sibling;
        full_path = svn_path_join(path, node->name, svn->p);
        sv_rss_print_changed_tree(node, full_path, svn);
    }

    return SVN_NO_ERROR;
}
